#'Compute persistence 
#'
#'Compute a persistence forecast based on a lagged autoregression of 
#'observational data along the time dimension, with a measure of forecast 
#'uncertainty (prediction interval) based on Coelho et al., 2004.\cr\cr
#'
#'@param data A numeric array corresponding to the observational data
#'  including the time dimension along which the autoregression is computed. The data
#'  should start at least 30 years before 'sdate_beg'.
#'@param time_dim A character string indicating the dimension along which to 
#'  compute the autoregression. The default value is 'time'.
#'@param sdate_beg A 4-digit integer indicating the first start date of the persistence
#'  forecast.
#'@param sdate_end A 4-digit integer indicating the last start date of the persistence
#'  forecast.
#'@param fyr_beg An integer indicating the forecast year for which
#'  the persistence forecast should be calculated, or the first forecast
#'  year of the average forecast years for which persistence should be
#'  calculated.
#'@param fyr_end An (optional) integer indicating the last forecast year
#'  of the average forecast years for which persistence should be calculated
#'  in the case of a multiyear average persistence. The default value is 'fyr_beg'.
#'@param memb An integer indicating the number of ensemble members to
#'  generate for the persistence forecast. The default value is 1.
#'@param ncores An integer indicating the number of cores to use for parallel
#'  computation. The default value is NULL.
#'
#'@return 
#'A list containing:
#'\item{$persistence}{
#'  A numeric array with dimensions 'memb', start dates, latitudes and longitudes
#'  containing the persistence forecast.   
#'}
#'\item{$persistence.mean}{
#'  A numeric array with same dimensions as 'persistence', except the 'memb' dimension
#'  which is of length 1, containing the ensemble mean persistence forecast.
#'}
#'\item{$persistence.predint}{
#'  A numeric array with same dimensions as 'persistence', except the 'memb' dimension
#'  which is of length 1, containing the prediction interval of the persistence forecast.
#'}
#'\item{$AR.slope}{
#'  A numeric array with same dimensions as 'persistence', except the 'memb' dimension
#'  which is of length 1, containing the slope coefficient of the autoregression.
#'}
#'\item{$AR.intercept}{
#'  A numeric array with same dimensions as 'persistence', except the 'memb' dimension
#'  which is of length 1, containing the intercept coefficient of the autoregression.
#'}
#'\item{$AR.lowCI}{
#'  A numeric array with same dimensions as 'persistence', except the 'memb' dimension
#'  which is of length 1, containing the lower value of the confidence interval of the
#'  autoregression.
#'}
#'\item{$AR.highCI}{
#'  A numeric array with same dimensions as 'persistence', except the 'memb' dimension
#'  which is of length 1, containing the upper value of the confidence interval of the
#'  autoregression.
#'}
#'
#'@examples
#'# Load sample data as in Load() example:
#'example(Load)
#'months_between_startdates <- 60
#'persist <- Persistence(sampleData$obs, sdate_beg = 1961, sdate_end = 2005, fyr_beg = 1, memb = 40)
#'
#'@rdname Persistence
#'@import multiApply
#'@export
Persistence <- function(data, time_dim = 'time', sdate_beg, sdate_end, fyr_beg, 
                  fyr_end = fyr_beg, memb = 1, ncores = NULL) {

  # Check inputs 
  ## data
  if (is.null(data)) {
    stop("Parameter 'data' cannot be NULL.")
  }
  if (!is.numeric(data)) {
    stop("Parameter 'data' must be a numeric array.")
  }
  if (is.null(dim(data))) {  #is vector
    dim(data) <- c(length(data))
    names(dim(data)) <- time_dim
  }
  if(any(is.null(names(dim(data))))| any(nchar(names(dim(data))) == 0)) {
    stop("Parameter 'data' must have dimension names.")
  }
  ## time_dim
  if (!is.character(time_dim) | length(time_dim) > 1) {
    stop("Parameter 'time_dim' must be a character string.")
  }
  if (!time_dim %in% names(dim(data))) {
    stop("Parameter 'time_dim' is not found in 'data' dimension.")
  }
  ## sdate_beg
  if (!is.numeric(sdate_beg) | sdate_beg %% 1 != 0 | sdate_beg < 0 |
      length(sdate_beg) > 1) | sdate_beg < 1850 | sdate_beg > 2020 {
    stop("Parameter 'sdate_beg' must be an integer between 1850 and 2020.")
  }
  ## sdate_end
  if (!is.numeric(sdate_end) | sdate_end %% 1 != 0 | sdate_end < 0 |
      length(sdate_end) > 1) | sdate_end < 1850 | sdate_end > 2020 {
    stop("Parameter 'sdate_end' must be an integer between 1850 and 2020.")
  }
  ## fyr_beg
  if (!is.numeric(fyr_beg) | fyr_beg %% 1 != 0 | fyr_beg < 0 |
      length(fyr_beg) > 1) {
    stop("Parameter 'fyr_beg' must be a positive integer.")
  }
  ## fyr_end
  if (!is.numeric(fyr_end) | fyr_end %% 1 != 0 | fyr_end < 0 |
      length(fyr_end) > 1) | fyr_end > 10 {
    stop("Parameter 'fyr_end' must be a positive integer below 10.")
  }
  ## memb
  if (!is.numeric(memb) | memb %% 1 != 0 | memb < 0 |
      length(memb) > 1) {
    stop("Parameter 'memb' must be a positive integer.")
  }
  ## ncores
  if (!is.null(ncores)) {
    if (!is.numeric(ncores) | ncores %% 1 != 0 | ncores < 0 |
      length(ncores) > 1) {
      stop("Parameter 'ncores' must be a positive integer.")
    }
  }

  ###############################
  # Calculate Trend
  dim_names <- names(dim(data))

  if (conf) {
    output_dims <- list(trend = 'stats', conf.lower = 'stats',
                        conf.upper = 'stats', detrended = time_dim)
  } else if (!conf) {
    output_dims <- list(trend = 'stats', detrended = time_dim)
  }


  output <- Apply(list(data),
                  target_dims = time_dim,
                  fun = .Trend,
                  output_dims = output_dims,
                  time_dim = time_dim, interval = interval, 
                  polydeg = polydeg, conf = conf,
                  conf.lev = conf.lev,
                  ncores = ncores)

  #output <- lapply(output, .reorder, time_dim = time_dim, dim_names = dim_names)

  return(output)
}

.Trend <- function(x, time_dim = 'sdate', interval = 1, polydeg = 1,
                   conf = TRUE, conf.lev = 0.95) {

  mon <- seq(x) * interval

  # remove NAs for potential poly()
  NApos <- 1:length(x)
  NApos[which(is.na(x))] <- NA
  x2 <- x[!is.na(NApos)]
  mon2 <- mon[!is.na(NApos)]

  if (length(x2) > 0) {
#    lm.out <- lm(x ~ mon, na.action = na.omit)
    lm.out <- lm(x2 ~ poly(mon2, degree = polydeg, raw = TRUE), na.action = na.omit)
    trend <- lm.out$coefficients  #intercept, slope1, slope2,...

    if (conf) {
      conf.lower <- confint(lm.out, level = conf.lev)[, 1]
      conf.upper <- confint(lm.out, level = conf.lev)[, 2]
    }

    detrended <- c()
    detrended[is.na(x) == FALSE] <- x[is.na(x) == FALSE] - lm.out$fitted.values
  } else {
    trend <- rep(NA, polydeg + 1)
    detrend <- NA
    if (conf) {
      conf.lower <- rep(NA, polydeg + 1)
      conf.upper <- rep(NA, polydeg + 1)
    }
  }

  if (conf) {
    return(list(trend = trend, conf.lower = conf.lower, conf.upper = conf.upper, 
                detrended = detrended))
  } else {
    return(list(trend = trend, detrended = detrended))
  }

}
